import type {
  Document,
  DirEntry,
  FileEntry,
  FUID,
  DirList,
  SelectedItems
} from '@/repositories/Document'
import { formatSize, formatUnixDate } from '@/utils'
import { defineStore } from 'pinia'
// @ts-ignore
import { localeIncludes } from 'locale-includes'

type FileData = { id: string; mtime: number; size: number; dir: DirectoryData }
type DirectoryData = {
  [filename: string]: FileData
}
type User = {
  username: string
  privileged: boolean
  isOpenLoginModal: boolean
  isLoggedIn: boolean
}

export type DocumentStore = {
  root: DirEntry
  document: Document[]
  selected: Set<FUID>
  uploadingDocuments: Array<{ key: number; name: string; progress: number }>
  uploadCount: number
  wsWatch: WebSocket | undefined
  wsUpload: WebSocket | undefined
  fileExplorer: any
  user: User
  error: string
}

export const useDocumentStore = defineStore({
  id: 'documents',
  state: (): DocumentStore => ({
    root: {} as DirEntry,
    document: [] as Document[],
    selected: new Set<FUID>(),
    uploadingDocuments: [],
    uploadCount: 0 as number,
    wsWatch: undefined,
    wsUpload: undefined,
    fileExplorer: null,
    error: '' as string,
    user: {
      username: '',
      privileged: false,
      isLoggedIn: false,
      isOpenLoginModal: false
    } as User
  }),

  actions: {
    updateTable(matched: DirList) {
      // Transform data
      const dataMapped = []
      for (const [name, attr] of Object.entries(matched)) {
        const { id, size, mtime } = attr
        const element: Document = {
          name,
          key: id,
          size,
          sizedisp: formatSize(size),
          mtime,
          modified: formatUnixDate(mtime),
          type: 'dir' in attr ? 'folder' : 'file'
        }
        dataMapped.push(element)
      }
      // Pre sort directory entries folders first then files, names in natural ordering
      dataMapped.sort((a, b) =>
        a.type === b.type
          ? a.name.localeCompare(b.name, undefined, {
              numeric: true,
              sensitivity: 'base'
            })
          : a.type === 'folder'
          ? -1
          : 1
      )
      this.document = dataMapped
    },
    setFilter(filter: string) {
      if (filter === '') return this.updateTable({})
      function traverseDir(data: DirEntry | FileEntry, path: string) {
        if (!('dir' in data)) return
        for (const [name, attr] of Object.entries(data.dir)) {
          const fullname = `${path}/${name}`
          if (
            localeIncludes(name, filter, {
              usage: 'search',
              numeric: true,
              sensitivity: 'base'
            })
          ) {
            matched[fullname.slice(1)] = attr // No initial slash on name
          }
          traverseDir(attr, fullname)
        }
      }
      const matched: any = {}
      traverseDir(this.root, '')
      this.updateTable(matched)
    },
    setActualDocument(location: string) {
      location = decodeURIComponent(location)
      let data: FileEntry | DirEntry = this.root
      const actualDirArr = []
      try {
        // Navigate to target folder
        for (const dirname of location.split('/').slice(1)) {
          if (!dirname) continue
          if (!('dir' in data)) throw Error('Target folder not available')
          actualDirArr.push(dirname)
          data = data.dir[dirname]
        }
      } catch (error) {
        console.error(
          'Cannot show requested folder',
          location,
          actualDirArr.join('/'),
          error
        )
      }
      if (!('dir' in data)) {
        // Target folder not available
        this.document = []
        return
      }
      this.updateTable(data.dir)
    },
    updateUploadingDocuments(key: number, progress: number) {
      for (const d of this.uploadingDocuments) {
        if (d.key === key) d.progress = progress
      }
    },
    pushUploadingDocuments(name: string) {
      this.uploadCount++
      const document = {
        key: this.uploadCount,
        name: name,
        progress: 0
      }
      this.uploadingDocuments.push(document)
      return document
    },
    deleteUploadingDocument(key: number) {
      this.uploadingDocuments = this.uploadingDocuments.filter(e => e.key !== key)
    },
    updateModified() {
      for (const d of this.document) {
        if ('mtime' in d) d.modified = formatUnixDate(d.mtime)
      }
    },
    login(username: string, privileged: boolean) {
      this.user.username = username
      this.user.privileged = privileged
      this.user.isLoggedIn = true
      this.user.isOpenLoginModal = false
    }
  },
  getters: {
    mainDocument(): Document[] {
      return this.document
    },
    isUserLogged(): boolean {
      return this.user.isLoggedIn
    },
    selectedFiles(): SelectedItems {
      function traverseDir(data: DirEntry | FileEntry, path: string, relpath: string) {
        if (!('dir' in data)) return
        for (const [name, attr] of Object.entries(data.dir)) {
          const fullname = path ? `${path}/${name}` : name
          // Is this the file we are looking for? Ignore if nested within another selection.
          let r = relpath
          if (selected.has(attr.id) && !relpath) {
            ret.selected.add(attr.id)
            ret.rootdir[name] = attr
            r = name
          } else if (relpath) {
            r = `${relpath}/${name}`
          }
          if (r) {
            ret.entries[attr.id] = attr
            ret.fullpath[attr.id] = fullname
            ret.relpath[attr.id] = r
            ret.ids.push(attr.id)
            if (!('dir' in attr)) ret.url[attr.id] = `/files/${fullname}`
          }
          traverseDir(attr, fullname, r)
        }
      }
      const selected = this.selected
      const ret: SelectedItems = {
        selected: new Set<FUID>(),
        missing: new Set<FUID>(),
        rootdir: {} as DirList,
        entries: {} as Record<FUID, FileEntry | DirEntry>,
        fullpath: {} as Record<FUID, string>,
        relpath: {} as Record<FUID, string>,
        url: {} as Record<FUID, string>,
        ids: [] as FUID[]
      }
      traverseDir(this.root, '', '')
      // What did we not select?
      for (const id of selected) {
        if (!ret.selected.has(id)) ret.missing.add(id)
      }
      // Sorted array of FUIDs for easy traversal
      ret.ids.sort((a, b) =>
        ret.relpath[a].localeCompare(ret.relpath[b], undefined, {
          numeric: true,
          sensitivity: 'base'
        })
      )
      return ret
    }
  }
})
